@
@ Makrobibliothek
@
@ (C) 27/11/90 Martin Patzel
@
@ Das Hauptprogramm, das diese Makrobibliothek benutzt, muss folgende
@ Variablen/Labels definieren:
@ macbuf	.word	0	; Pointer fuer Makroverwaltung
@ end		EQU	.	; Letztes Label! Danach kommt freier Speicher!
@
@ Initialisierung der Makrobibliothek:
@		InitMem		; Speicherverwaltung initialisieren
@		InitSP	#0x100	; Stack anlegen
@		InitMac		; Makro-Speicher anlegen
@


.set MACBUFSIZE__,	0x10
.set BranchAlways,	0xEA000000	@ Maschinencode fuer B - Befehl



@=======================================================================
@============== Makros zur Erweiterung des Befehlssatzes ===============
@=======================================================================

@
@ MACRO		NOP
@
@ Pseudo-Opcode
@
		.macro		NOP
\label		MOVNV	R0, R0
		.endm

@
@ MACRO		LDIND	$reg, $addr
@
@ laedt das Register \reg mit [\addr]
@
		.macro		LDIND	reg, addr
\label		LDR	\reg, p\addr
		LDR	\reg, [\reg, #0]
		.endm


@=======================================================================
@============== Sonstige Makros ========================================
@=======================================================================

@
@ MACRO		GetLog2	$from, $to
@
@ \to = log2(\from)
@ \from ist modifiziert!
@
		.macro		GetLog2	from, to
		ROUT
\label		MOV	\to, #0
0:		MOVS	\from, \from, LSR #1
		ADDCC	\to, \to, #1
		BCC	0b_or_f
		ROUT
		.endm


@
@ MACRO		GenBra	$fromreg, $destreg
@
@ Erzeugt einen Sprungbefehl von Adresse \fromreg, der zur Adresse
@ \destreg springt. In \destreg wird der Sprungbefehl zurueckgeliefert.
@ in \fromreg hat die absolute Adresse, wo der B-Befehl hingeladen
@ werden soll, zu stehen.
@
		.macro		GenBra	fromreg, destreg
\label		SUB	\destreg, \destreg, \fromreg
		MOV	\destreg, \destreg, LSR #2	@ Sprungweite in Worten!
		SUB	\destreg, \destreg, #2		@ PC-Prefetch korrigieren
		ORR	\destreg, \destreg, #BranchAlways
		.endm


@
@ MACRO		SetVec	$VecNoreg, $newreg, $oldreg
@
@ Aendert den Vektor in Adresse \VecNoreg so ab, dass er auf die Adresse
@ \newreg zeigt. Das alte Sprungziel wird in \oldreg zurueckgeliefert.
@
		.macro		SetVec	VecNoreg, newreg, oldreg
\label		GenBra	\VecNoreg, \newreg
		LDR	\oldreg, [\VecNoreg, #0]     @ alten Vektor holen
		BIC	\oldreg, \oldreg, #0xFF000000 @ Wert aus B-Befehl holen
		ADD	\oldreg, \oldreg, #2	     @ Prefetch-Korrektur
		MOV	\oldreg, \oldreg, LSL #2     @ war in Worten!
		ADD	\oldreg, \oldreg, \VecNoreg  @ Zieladresse!
		STR	\newreg, [\VecNoreg, #0]     @ neuen Vektor setzen
		.endm


@
@ MACRO		TestPat $max
@
@ Belegt die Register R0-R\max mit einem Testpattern
@
		.macro		TestPat	max
\label		[ \max >= 0
		MOV	R0, #0x00
		]
		[ \max >= 1
		MOV	R1, #0x11
		]
		[ \max >= 2
		MOV	R2, #0x22
		]
		[ \max >= 3
		MOV	R3, #0x33
		]
		[ \max >= 4
		MOV	R4, #0x44
		]
		[ \max >= 5
		MOV	R5, #0x55
		]
		[ \max >= 6
		MOV	R6, #0x66
		]
		[ \max >= 7
		MOV	R7, #0x77
		]
		[ \max >= 8
		MOV	R8, #0x88
		]
		[ \max >= 9
		MOV	R9, #0x99
		]
		[ \max >= 10
		MOV	R10, #0xAA
		]
		[ \max >= 11
		MOV	R11, #0xBB
		]
		[ \max >= 12
		MOV	R12, #0xCC
		]
		[ \max >= 13
		MOV	R13, #0xDD
		]
		[ \max >= 14
		MOV	R14, #0xEE
		]
		.endm


@=======================================================================
@==================== Memoryverwaltung =================================
@=======================================================================

@
@ MACRO		InitMem
@
@ Initialisiert Memoryverwaltung
@
@ Uses R0-R3, Label end
@
		.macro		InitMem
\label		SWI	OS_GetEnv
		ADRL	R0, end
		ADD	R0, R0, #3
		AND	R0, R0, #0xFFFFFFFC	@ auf Wortgrenze gehen
		SUB	R3, R1, R0		@ R3 = # freie Bytes
		MOV	R1, R0
		MOV	R0, #0			@ Heap initialisieren
		SWI	OS_Heap
		NOP
		.endm


@
@ MACRO		InitSP	$size
@
@ Initialisiert Stackpointer; zuvor muss mit InitMem die Heapverwaltung
@ initialisiert worden sein. Stack wird in der Groesse \size angelegt
@
@ Uses R0-R3, SP, Label end
@
		.macro		InitSP	size
\label		ADRL	R1, end
		MOV	R0, #2
		MOV	R3, \size
		SWI	OS_Heap			@ Stack anfordern
		NOP
		ADD	SP, R2, \size		@ SP initialisieren
		.endm


@
@ MACRO		InitMac
@
@ Initialisiert Speicher fuer Makros
@
@ Uses R0-R3, Label end, macbuf
@
		.macro		InitMac
\label		ADRL	R1, end
		MOV	R0, #2
		MOV	R3, #MACBUFSIZE__
		SWI	OS_Heap			@ Puffer anfordern
		STR	R2, macbuf
		.endm



@
@ MACRO		MAlloc	$reg, $size
@
@ Fordert Speicher an, #Bytes steht in \size, Pointer auf
@ Speicherbereich steht nachher in \reg, \reg = 0, falls
@ Speicher nicht beschafft werden konnte
@
@ Uses: Label end
@
		.macro		MAlloc	reg, size
\label		STMDB	SP!, {R0-R3}	@ Register sichern
		MOV	R3, \size
		ADRL	R1, end
		MOV	R0, #2
		SWI	OS_Heap+XBit	@ Speicher holen
		LDR	R0, macbuf	@ Adresse zwischenspeichern
		STR	R2, [R0]
		LDMIA	SP!, {R0-R3}	@ Register restaurieren
		LDR	\reg, macbuf
		LDR	\reg, [\reg]	@ Adresse in Zielregister holen
		.endm


@
@ MACRO		MFree	$adr
@
@ Gibt Speicher an Adresse \adr wieder frei
@
@ Uses: Label end
@
		.macro		MFree	adr
\label		STMDB	SP!, {R0-R2}
		MOV	R2, \adr
		ADRL	R1, end
		MOV	R0, #3
		SWI	OS_Heap+XBit
		NOP
		LDMIA	SP!, {R0-R2}
		.endm


@
@ MACRO		MReAlloc $reg, $adr, $size
@
@ Vergroessert (\size > 0) bzw. verkleinert (\size < 0) den
@ Speicherblock, der bei \adr steht um den in \size angegebenen Betrag.
@ Adresse auf neuen Speicherblock erhaelt man in \reg
@
@ Uses: Label end
@
		.macro		MReAlloc reg, adr, size
\label		STMDB	SP!, {R0-R3}	@ Register sichern
		MOV	R3, \size
		MOV	R2, \adr
		ADRL	R1, end
		MOV	R0, #4
		SWI	OS_Heap+XBit	@ Realloc durchfuehren
		LDR	R0, macbuf	@ Adresse zwischenspeichern
		STR	R2, [R0]
		LDMIA	SP!, {R0-R3}	@ Register restaurieren
		LDR	\reg, macbuf
		LDR	\reg, [\reg]	@ Adresse in Zielregister holen
		.endm


@=======================================================================
@============== Druck-Makros ===========================================
@=======================================================================

@
@ MACRO		SPrint	$str
@
@ Druckt den String \str am Bildschirm aus
@

		.macro		SPrint	str
		ROUT
\label		STMDB	SP!, {R0,LK}
		ADR	R0, 0b_or_f
		SWI	OS_Write0
		B	1b_or_f
0:
		=	"\str"
		=	0
		.balign 4
1:		LDMIA	SP!, {R0,LK}
		ROUT
		.endm

@
@ MACRO		HPrint	$reg
@
@ Druckt den Inhalt von Register \reg in Hex am Bildschirm aus
@
		.macro		HPrint	reg
\label		STMDB	SP!,{R0-R2, LK}
		MOV	R0, \reg
		LDR	R1, macbuf
		MOV	R2, #MACBUFSIZE__
		SWI	OS_ConvertHex8
		SWI	OS_Write0
		NOP
		LDMIA	SP!,{R0-R2, LK}
		.endm

@
@ MACRO		NLPrint
@
@ Beginnt eine neue Zeile am Bildschirm
@
		.macro		NLPrint
\label		STMDB	SP!, {LK}
		SWI	OS_NewLine
		NOP
		LDMIA	SP!, {LK}
		.endm

@
@ MACRO		RPrint	$str, $reg
@
@ Druckt den String \str aus, dann den Inhalt des Registers \reg
@ und dann noch einen Zeilenvorschub.
@
		.macro		RPrint	str, reg
\label		SPrint	"\str"
		HPrint	\reg
		NLPrint
		.endm

@
@ MACRO		RDump
@
@ Druckt alle Register aus
@
		.macro		RDump
\label		RPrint	"R0 =", R0
		RPrint	"R1 =", R1
		RPrint	"R2 =", R2
		RPrint	"R3 =", R3
		RPrint	"R4 =", R4
		RPrint	"R5 =", R5
		RPrint	"R6 =", R6
		RPrint	"R7 =", R7
		RPrint	"R8 =", R8
		RPrint	"R9 =", R9
		RPrint	"R10=", R10
		RPrint	"R11=", R11
		RPrint	"R12=", R12
		RPrint	"R13=", R13
		RPrint	"R14=", R14
		.endm

@
@ MACRO		FieldPrint $field,$num
@
@ Druckt den Inhalt des Feldes \field (\num Worte gross) aus
@
		.macro		FieldPrint	field, num
\label		STMFD	SP!, {R0-R3}
		ADRL	R0, \field
		LDR	R1, =\num
		MOV	R2, #0
l0:		LDR	R3, [R0, R2, LSL #2]
		SPrint	"&"
		HPrint	R3
		SPrint	", "
		ADD	R2, R2, #1
		TST	R2, #7
		BNE	l1
		NLPrint
l1:		CMP	R2, R1
		BLT	l0
		LDMFD	SP!, {R0-R3}
		.endm

@
@ MACRO		R0FieldPrint $num
@
@ Druckt den Inhalt des Feldes, auf das R0 zeigt, (\num Worte gross) aus
@
		.macro		R0FieldPrint	num
\label		STMFD	SP!, {R0-R3}
		LDR	R1, =\num
		MOV	R2, #0
l0:		LDR	R3, [R0, R2, LSL #2]
		SPrint	"&"
		HPrint	R3
		SPrint	", "
		ADD	R2, R2, #1
		TST	R2, #7
		BNE	l1
		NLPrint
l1:		CMP	R2, R1
		BLT	l0
		LDMFD	SP!, {R0-R3}
		.endm


		@ END
